#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""
@File Name  : localhost_mysql.py
@Author     : LeeCQ
@Date-Time  : 2019/12/6 18:40

本地SQL测试文件
"""


from sqllib.mysql import MyMySqlAPI
# from DBUtils.PooledDB import PooledDB


class LocalhostMySQL(MyMySqlAPI):
    """"""

    def __init__(self, user, passwd, db, **kwargs):
        super().__init__('localhost', 3306, user, passwd, db, 'utf8', **kwargs)
        self.table_name = 'insert_multiple_at_once'

    def test_create_table(self):
        _c = (f"CREATE TABLE IF NOT EXISTS `{self.table_name}` ( "
              f"a VARCHAR(10),"
              f"b VARCHAR(10)"
              f" ) ")
        super()._create_table(_c)

    def test_insert(self):
        """"""
        self.test_create_table()
        a = 'a1', 'a2'
        b = 'b1', 'b2'
        a = 'aa'
        return self._insert(self.table_name, a=a, b=b)


if __name__ == '__main__':
    # 忽略警告；
    import warnings
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        LocalhostMySQL('root', '123456', 'test').test_insert()



#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""
@File Name  : connection_pool.py
@Author     : LeeCQ
@Date-Time  : 2019/12/15 16:51

"""
from dbutils.pooled_db import PooledDB

dbs = ''

POOL = PooledDB(
    creator=dbs,   # 使用链接数据库的模块
    maxconnections=6,  # 连接池允许的最大连接数，0和None表示不限制连接数
    mincached=2,  # 初始化时，链接池中至少创建的空闲的链接，0表示不创建
    maxcached=5,  # 链接池中最多闲置的链接，0和None不限制
    maxshared=3,  # 链接池中最多共享的链接数量，0和None表示全部共享。
    # PS: 无用，因为pymysql和MySQLdb等模块的 threadsafety都为1，所有值无论设置为多少，_maxcached永远为0，所以永远是所有链接都共享。
    blocking=True,  # 连接池中如果没有可用连接后，是否阻塞等待。True，等待；False，不等待然后报错
    maxusage=None,  # 一个链接最多被重复使用的次数，None表示无限制
    setsession=[],  # 开始会话前执行的命令列表。如：["set datestyle to ...", "set time zone ..."]
    ping=1,  # ping MySQL服务端，检查是否服务可用，
    # 如：0 = None = never, 1 = default = whenever it is requested, 2 = when a cursor is created,
    # 4 = when a query is executed, 7 = always
    host="127.0.0.1",
    port=3306,
    user="root",
    password="123456",
    database="test",
    )


def func():
    # 检测当前正在运行连接数的是否小于最大链接数，如果不小于则：等待或报raise TooManyConnections异常
    # 否则
    # 则优先去初始化时创建的链接中获取链接 SteadyDBConnection。
    # 然后将SteadyDBConnection对象封装到PooledDedicatedDBConnection中并返回。
    # 如果最开始创建的链接没有链接，则去创建一个SteadyDBConnection对象，再封装到PooledDedicatedDBConnection中并返回。
    # 一旦关闭链接后，连接就返回到连接池让后续线程继续使用。
    conn = POOL.connection()
    cursor = conn.cursor()
    cursor.execute('select * from 庆余年')
    result = cursor.fetchall()
    print(result)
    conn.close()


func()



#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""
@Time   : 2019/11/26 12:54
@Author : LeeCQ
@File Name: __init__.py


"""

from .mysqlbase import MyMySqlAPI, MySqlAPI


class Test(MyMySqlAPI):
    """测试的mysql对象封装"""

    def __init__(self, host='localhost', port=3306,
                 user='test', passwd='test123456', db='test', charset='gb18030',
                 **kwargs):
        super().__init__(host, port, user, passwd, db, charset, **kwargs)



#!/usr/bin/python3 
# -*- coding: utf-8 -*-
"""
@File Name  : __init__.py
@Author     : LeeCQ
@Date-Time  : 2020/1/13 12:30

SQLite

2020.01.28 -- _select OFFSET 添加str()
"""

from .sqlite import SQLiteAPI



#!/usr/bin/python3 
# -*- coding: utf-8 -*-
"""
@File Name  : __init__.py
@Author     : LeeCQ
@Date-Time  : 2020/1/13 14:22

https://www.runoob.com/python3/python-mongodb.html
https://www.runoob.com/mongodb/mongodb-tutorial.html

ORM 框架
"""
import pymongo

client = pymongo.MongoClient("mongodb+srv://root:root123456@test-ivnj5.azure.mongodb.net/test?retryWrites=true&w=majority")
db = client.test
col = db['book']

col.insert_one()


#!/usr/bin/python3 
# -*- coding: utf-8 -*-
"""
@File Name  : __init__.py
@Author     : LeeCQ
@Date-Time  : 2020/1/13 17:15

https://blog.csdn.net/langkew/article/details/48480069 Berkeley DB

Berkeley DB
Berkeley DB是一个纯Java的开源的文件数据库，介于关系数据库与内存数据库之间，使用方式与内存数据库类似，
它提供的是一系列直接访问数据库的函数，而不是像关系数据库那样需要网络通讯、SQL解析等步骤。

爬虫框架WebCollector就用到了BerkeleyDB。pip
"""

import bsddb3

db = bsddb3.db
db = db.DB()



#!/usr/bin/python3 
# -*- coding: utf-8 -*-
"""
@File Name  : game_num.py
@Author     : LeeCQ
@Date-Time  : 2020/1/13 12:30

https://plugins.jetbrains.com/pluginManager/?action=download&id=DBN&build=PC-191.8026.44&uuid=a59f1549-d4cd-4d46-ab22-be8a4837bcb7
"""
import sqlite3

conn = sqlite3.connect('./sup/test.db')


books = [(1, 1, 'Cook Recipe', 3.12, 1),
         (2, 3, 'Python Intro', 17.5, 2),
         (3, 2, 'OS Intro', 13.6, 2),
         ]

cur = conn.cursor()

# create tables
cur.execute('''CREATE TABLE IF NOT  EXISTS category
      (id int primary key, sort int, name text)''')
cur.execute('''CREATE TABLE IF NOT  EXISTS book 
      (id int primary key,
       sort int,
       name text,
       price real,
       category int,
       FOREIGN KEY (category) REFERENCES category(id))''')

# save the changes
conn.commit()

# close the connection with the database
cur.close()


cur = conn.cursor()
# execute "INSERT"
cur.execute("INSERT INTO category VALUES (1, 1, 'kitchen')")

# using the placeholder
cur.execute("INSERT INTO category (id , sort, name)VALUES (?, ?, ?)", (3, 2, 'computer'))

# execute multiple commands
cur.executemany('INSERT INTO book VALUES (?, ?, ?, ?, ?)', books)

conn.commit()
conn.close()






#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""
@File Name  : learn_unittest
@Author     : LeeCQ
@Date-Time  : 2021/1/10 0:04
"""
import unittest

if __name__ == '__main__':
    # 第一种加载方式：
    # unittest.main()
    # 第二种加载方式：
    # 构建测试套件：将要执行的测试用例添加到测试套件里
    suite = unittest.TestSuite()
    suite.addTest(TestAdd("test_add1"))
    suite.addTest(TestSub("test_sub1"))
    suite.addTest(TestSub("test_sub2"))
    suite.addTest(TestAdd("test_add2"))
    # 通过测试运行器执行测试套件
    runner = unittest.TextTestRunner()
    runner.run(suite)

    # 使用unitest提供的discover方法自动发现测试用例

    # 定义测试用例的目录
    # test_dir='./'
    # #使用discover方法去发现测试用例放到discover里
    # discover=unittest.defaultTestLoader.discover(test_dir,pattern='test*.py')
    # #使用测试运行器去执行加载在discover里的测试用例
    # runner=unittest.TextTestRunner()
    # runner.run(discover)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""
@File Name  : SQLCommon
@Author     : LeeCQ
@Date-Time  : 2020/2/23 16:43

"""
import logging
import sys

logger = logging.getLogger("SQL")  # 创建实例
formatter = logging.Formatter("[%(asctime)s] < %(funcName)s: %(lineno)d > [%(levelname)s] %(message)s")
# 终端日志
console_handler = logging.StreamHandler(sys.stdout)
console_handler.setFormatter(formatter)  # 日志文件的格式
logger.setLevel(logging.DEBUG)  # 设置日志文件等级


def _get(_l: list or tuple, n: int) -> str:
    """列表中取值，"""
    try:
        return _l[n]
    except:
        return ''


def sql_join(data: (tuple, list)) -> tuple:
    """在创建表的时候，通过给定的元组生成SQL语句

    给定的元组形式：
        (字段名，共有属性，MySql属性，Sqlite属性)  -- 4 元素

    :param data: 这是嵌套元组 (( ), ...)
    :return (mysql, sqlite)
    """
    return (
        ', '.join(
            [_ for _ in [' '.join(
                [_i for _i in [m[0], m[1], _get(m, 2)] if _i.strip() != '']
            ) for m in data] if _.strip() != '']
        ),
        ', '.join(
            [_ for _ in [' '.join(
                [_i for _i in [m[0], m[1], _get(m, 3)] if _i.strip() != '']
            ) for m in data] if _.strip() != '']
        )
    )


class SQLiteJson:
    """SQLite的JSON数据类型支持"""


if __name__ == '__main__':
    _c = (('No', ' ', 'INT(4) PRIMARY KEY AUTO_INCREMENT', 'INTEGER PRIMARY KEY AUTOINCREMENT'),
          ('orderId', 'BIGINT UNIQUE'),
          ('orderNumber', 'BIGINT UNIQUE'),
          ('user', 'VARCHAR(20)'),
          ('passwd', 'VARCHAR(36)'),
          ('albumId', 'INT(5)'),
          ('albumName', 'VARCHAR(50)'),
          ('officePrice', 'FLOAT')
          )
    print(sql_join(_c)[0])



#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""
@Time   : 2019/11/26 12:54
@Author : LeeCQ
@File Name: mysql.py

更多操作详见 下方import 内容。
内容：
    * v1:
        1. 构建 class:MySQL 框架
            以 MySQL._write_db(), MySQL._write_rows(), MySQL._read_db()为基础访问SQL。
            创建_select, _insert, _update, _drop, _delete为基础的接口访问。
        2. 构建访问控制 & 安全 相关的语句。
            _key_and_table_is_exists
        3. 优化流程控制。
        4. 优化数据库访问。
    * v2: -2019/12/18
        1. 新增 DBUtils.PooledDB 模块：连接池
            1.1. 新增MyMySQL.pooled_sql()模块，以启用连接池
            1.2. 修改MyMySQL._write_db(), MySQL._write_rows(), MySQL._read_db():
                    当他的子类或者实例调用 -> MySQL.pooled_sql() <- 方法时，以开启连接池；
                    if self.pooled_sql is not None:
                        _sql = self.pooled_sql.connection()
                    else:
                        _sql = self._sql
        2. 微调 MySQL._create_table()方法：
            源：( with self._sql.cursor() as cur: \\ cur.execute(command, args) \\self._sql.commit() \\ return 0) ==>
            修改为：( return self._write_db(command, args) )
            有点：便于代码的重用性；

    * v3: -2020/01/11
        1. 新增 def _alter_table():  --> 向已有数据表中添加 键
        2. 更新详细了注释
        3. MySQL - 增加了对 pymysql.connect() - 所需参数的详细注解 -- 1/12

    *v4: - 2020/02/29
        1. 修改show_columns() - 表名前缀问题 <-  ok
        2. 修改show_table -> show_tables                     --ok  2020/07/20
        3. 修改show_tables逻辑 str(_c[0]) -> _c[0].decode()   -- ok 2020/07/20

"""

import logging
import sys
import pymysql
from sqllib.common.base_sql import BaseSQL, BaseSQLAPI
from sqllib.common.error import *
# from sqllib.common.common import sql_join
from dbutils.pooled_db import PooledDB
from warnings import filterwarnings

logger = logging.getLogger("mysql")  # 创建实例
formatter = logging.Formatter("[%(asctime)s] < %(funcName)s: %(lineno)d > [%(levelname)s] %(message)s")
# 终端日志
terminal_handler = logging.StreamHandler(sys.stdout)
terminal_handler.setFormatter(formatter)  # 日志文件的格式
logger.setLevel(logging.DEBUG)  # 设置日志文件等级

_all_ = ['MyMySqlAPI', 'MySqlAPI']


class MyBaseSQL(BaseSQL):
    """mysql 操作的模板：

    这个类包含了最基本的MySQL数据库操作，SELECT, INSERT, UPDATE, DELETE

    主要的接口方法：
    :method _select(self, table, column_name, *args, **kwargs):
            从数据库查询数据： - 表名，查询列，拓展查询列（可以使用通配符'*'）
                             - kwargs: {WHERE, LIMIT, OFFSET}  LIMIT 有一种新用法 - limit 偏移量，限制值
        :return 查找的数据

    :method _insert(self, table, **kwargs)
            向数据库插入新数据 - kwargs: 需要插入的数据的 键值对；
        :return 0 or -1

    :method _update(self, table, where_key, where_value, **kwargs)
            向数据库更新字段的值，  - where代表着查询的行  -- 如果没有将会更新所有行。
                                 - kwargs 插入的键值对
        :return 0 or -1

    :method _drop(self, option, name)：
            删除数据表或者数据库      - option = table or database
                                   - name 对应的名称
        :return 0 or -1

    :method _delete(self, table, where_key, where_value, **kwargs):

    :method write_db(self, command):
            没有模板时使用     - command: 数据库查询字符串。
        :return 0 or -1

    :method read_db(self, command):
                没有模板时使用     - command: 数据库查询字符串。
        :return 查询结果；


    :param str host:    链接的数据库主机；
    :param int port:    数据库服务器端口
    :param str user:    数据库用户名
    :param str passwd:  数据库密码
    :param str db:      数据库的DataBase
    :param str charset: 数据库的字符集
    :param str prefix:  表前缀
    """

    def __init__(self, host, port, user, passwd, db, charset,
                 use_unicode=None, pool=False, **kwargs):
        super().__init__()
        self.SQL_HOST = host  # 主机
        self.SQL_PORT = port  # 端口
        self.SQL_USER = user  # 用户
        self.SQL_PASSWD = passwd  # 密码
        self.SQL_DB = db  # 数据库
        self.SQL_CHARSET = charset  # 编码
        self.use_unicode = use_unicode
        # 表前缀
        self.TABLE_PREFIX = '' if 'prefix' not in kwargs.keys() else kwargs['prefix']
        self._sql = pymysql.connect(host=self.SQL_HOST,
                                    port=self.SQL_PORT,
                                    user=self.SQL_USER,
                                    password=self.SQL_PASSWD,  # 可以用 passwd为别名
                                    database=self.SQL_DB,  # 可以用 db    为别名；
                                    charset=self.SQL_CHARSET,
                                    use_unicode=use_unicode,
                                    **kwargs
                                    )
        self.pooled_sql = None
        self.pooling_sql() if pool else None

    def set_use_db(self, db_name):
        """设置当前数据库"""
        return self._sql.select_db(db_name)

    def set_charset(self, charset):
        """设置数据库链接字符集"""
        return self._sql.set_charset(charset)

    # 建立连接池
    def pooling_sql(self, min_cached=0, max_cached=0,
                    max_connections=10, blocking=True,
                    max_usage=None, set_session=None, reset=True,
                    failures=None, ping=1,
                    **kwargs):
        """ 连接池建立

       ::param creator: 数据库连接池返回的模块；默认pymysql
        :param min_cached: 初始化时，池中最小链接数；
        :param max_cached: 链接池中最多闲置的链接，0和None不限制；
        :param max_connections: 池中最大链接数；
        :param blocking: 链接数用尽时，是否阻塞等待链接 True 等待 -- False 不等待 & 报错
        :param max_usage: 一个链接最多被重复使用的次数，None表示无限制
        :param set_session: # 开始会话前执行的命令列表。如：["set datestyle to ...", "set time zone ..."]
        :param reset: 当连接返回到池中时，应该如何重置连接
        :param failures:
        :param ping: ping MySQL服务端，检查是否服务可用，
        :param kwargs: {host=, poet=, user=, passwd=, database=, charset=}
        """
        self.pooled_sql = PooledDB(creator=pymysql, mincached=min_cached, maxcached=max_cached,
                                   maxconnections=max_connections, blocking=blocking,
                                   maxusage=max_usage, setsession=set_session, reset=reset,
                                   failures=failures, ping=ping,
                                   host=self.SQL_HOST if 'host' not in kwargs.keys() else kwargs['host'],
                                   port=self.SQL_PORT if 'post' not in kwargs.keys() else kwargs['port'],
                                   user=self.SQL_USER if 'user' not in kwargs.keys() else kwargs['user'],
                                   password=self.SQL_PASSWD if 'passwd' not in kwargs.keys() else kwargs['passwd'],
                                   database=self.SQL_DB if 'db' not in kwargs.keys() else kwargs['db'],
                                   charset=self.SQL_CHARSET if 'charset' not in kwargs.keys() else kwargs['charset']
                                   )

    def set_prefix(self, prefix):
        """设置表前缀"""
        self.TABLE_PREFIX = prefix

    def close(self):
        """关闭数据库连接"""
        self._sql.close()

    def _write_db(self, command, args=None):
        """执行数据库写入操作

        :type args: str, list or tuple
        """
        if self.pooled_sql is not None:
            _sql = self.pooled_sql.connection()
        else:
            _sql = self._sql

        cur = _sql.cursor()  # 使用cursor()方法获取操作游标
        try:
            _c = cur.execute(command, args)
            _sql.commit()  # 提交数据库
            return _c
        except Exception:
            _sql.rollback()
            sys.exc_info()
            raise SqlWriteError(f'操作数据库时出现问题，数据库已回滚至操作前——\n{sys.exc_info()}\n\n{command}')
        finally:
            cur.close()

    # 写入事务
    def _write_affair(self, command, args):
        """向数据库写入多行"""
        if self.pooled_sql is not None:
            _sql = self.pooled_sql.connection()
        else:
            _sql = self._sql

        try:
            with _sql.cursor() as cur:  # with 语句自动关闭游标
                _c = cur.executemany(command, args)
                _sql.commit()
            return _c
        except Exception:
            _sql.rollback()
            sys.exc_info()
            raise SqlWriteError("_write_rows() 操作数据库出错，已回滚 \n" + str(sys.exc_info()))

    def _read_db(self, command, args=None, result_type=None):
        """执行数据库读取数据， 返回结果

        :param result_type: 返回的结果集类型{dict, None, tuple, 'SSCursor', 'SSDictCursor'}
        """
        if self.pooled_sql is not None:
            _sql = self.pooled_sql.connection()
        else:
            _sql = self._sql

        ret_ = {dict: pymysql.cursors.DictCursor,
                None: pymysql.cursors.Cursor,
                tuple: pymysql.cursors.Cursor,
                list: pymysql.cursors.Cursor,
                'SSCursor': pymysql.cursors.SSCursor,
                'SSDictCursor': pymysql.cursors.SSDictCursor
                }
        cur = _sql.cursor(ret_[result_type])
        cur.execute(command, args)
        results = cur.fetchall()
        cur.close()
        return results

    # 查表中键的所有信息 - > list
    def _columns(self, table, result_type=None):
        """返回table中列（字段）的所有信息

         +-------+-------+------+------+-----+---------+-------+
         | index |   0   |  1   |   2  |  3  |    4    |   5   |
         +-------+-------+------+------+-----+---------+-------+
         | dict  | Field | Type | Null | Key | Default | Extra |
         +-------+-------+------+------+-----+---------+-------+
        """
        table = self.get_real_table_name(table)
        return self._read_db(f'show columns from `{table}`', result_type=result_type)

    # 查表中的键
    def columns_name(self, table) -> list:
        """返回 table 中的 列名在一个列表中"""
        table = self.get_real_table_name(table)
        return [_c[0].decode() if isinstance(_c[0], bytes) else _c[0] for _c in self._columns(table)]

    # 获取数据库的表名
    def tables_name(self) -> list:
        """由于链接时已经指定数据库，无需再次指定。返回数据库中所有表的名字。"""
        return [_c[0].decode() if isinstance(_c[0], bytes) else _c[0] for _c in self._read_db("show tables")]

    def _create_table(self, command: str, table_name, exists_ok=False, table_args='', *args):
        """回退强制要求传入 table_name"""

        if command.strip().endswith(','):
            command = command.strip()[:-1] + ' '

        _c = (f"CREATE TABLE {'IF NOT EXISTS' if exists_ok else ' '} "
              f"`{self.get_real_table_name(table_name)}` ( "
              + command +
              ") " + table_args)
        logger.debug(_c)
        # print(_c)
        return self._write_db(_c, args)

    # 插入表
    def _insert(self, table, ignore_repeat=False, **kwargs):
        """ 向数据库插入内容。

        :param table: 表名；
        :param ignore: 重复是否抛出异常
        :param kwargs: 字段名 = 值；
        :return:
        """
        ignore_ = 'IGNORE' if ignore_repeat else ''
        _c = (f"INSERT {ignore_} INTO `{self.get_real_table_name(table)}`  "
              "( " +
              ', '.join([" `" + _k + "` " for _k in kwargs.keys()]) +
              " ) "  # 这一行放在后面会发生，乱版；
              " VALUES "
              " ( " + ', '.join([" %s " for _k in kwargs.values()]) + " ) ; "  # 添加值
              )
        # print( _c)
        if not isinstance(list(kwargs.values())[0], (str, int, type(None), float)):
            arg = self.zip_data_for_insert(tuple(kwargs.values()))
            return self._write_affair(_c, arg)
        else:
            for x in kwargs.values():
                if isinstance(x, (list, tuple)):
                    raise InsertZipError("INSERT一条数据时，出现列表列或元组！确保数据统一")
            return self._write_db(_c, list(kwargs.values()))  # 提交

    def _insert_rows(self, table_name, args, k=None, ignore_repeat=False):
        """插入

        :param table_name:
        :param args:
        :param ignore_repeat:
        :return:
        """
        _a = [_.values() for _ in args]
        if k is None:
            if not isinstance(args[0], dict):
                raise ValueError(f'既没有k, 也不是dict')
            k = args[0].keys()
        _ignore = 'OR IGNORE' if ignore_repeat else ''
        _c = f"INSERT {_ignore} INTO {self.get_real_table_name(table_name)} ( "
        _c += ", ".join([_ for _ in k]) + " ) "
        _c += "VALUES ( "
        _c += ", ".join([f" %s " for _ in args[0].values()]) + ");"
        return self._write_affair(_c, _a)

    # 检索表
    def _select(self, table, columns_name: tuple and list, result_type=None, **kwargs):
        """ select的应用。

            ·· `就不再支持 * `
            ·· column_name, table, key 可以用 ` ` 包裹， value 一定不能用， value 用 ' ' 。
        :param result_type: {dict, None, tuple, 'SSCursor', 'SSDictCursor'}
        :param table:
        :param cols: 传参时自行使用 `` , 尤其是数字开头的参数
        :param kwargs: {'WHERE', 'LIMIT', 'OFFSET', ORDER} 全大写
        :return 结果集
        """

        command = f"SELECT  "
        command += ' , '.join(columns_name) + " "
        command += f'FROM `{self.get_real_table_name(table)}` '
        for key, value in kwargs.items():
            key = key.upper()
            if key in ['WHERE', 'LIMIT', 'OFFSET']:
                command += f' {key}  {value}'
            if key == 'ORDER':
                command += f' {key} BY {value}'
        # print(command, )
        return self._read_db(command, result_type=result_type)

    # 更新表
    def _update(self, table, where_key, where_value, **kwargs):
        """ 更新数据库 --2019/12/15

        :param table: 数据表名字
        :param where_key: where的键
        :param where_value: where的值
        :param kwargs: 更新的键 = 更新的值， 注意大小写，数字键要加 - ``
        :return: 0 成功。
        """
        self.key_and_table_is_exists(f'{self.get_real_table_name(table)}', where_key, **kwargs)  # 判断 表 & 键 的存在性！
        _update_data = ' , '.join([f" `{k}`=%({k})s  " for k, v in kwargs.items()])  # 构造更新内容
        command = (f"UPDATE `{self.get_real_table_name(table)}` SET  "
                   f"{_update_data}"
                   f" WHERE {where_key}='{where_value}' ;"  # 构造WHERE语句
                   )
        return self._write_db(command, kwargs)  # 执行SQL语句

    # 删除表或者数据库
    def _drop(self, option, name):
        """ 删除数据库内容：

        :param option: (TABLE or DATABASE)
        :param name:
        :return: 0 成功
        """
        if option.upper() == 'TABLE':
            command = f'DROP  {option}  `{self.get_real_table_name(name)}`'
        else:
            command = f'DROP  {option}  `{name}`'
        return self._write_db(command)

    def _delete(self, table, where_key, where_value, **kwargs):
        """删除数据表中的某一行数据，
        :param table:
        :param where_key: 最好是数据库的主键或唯一的键。如果数据库没有，则最好组合where，以保证删除 - 唯一行。
        :param where_value:
        :param kwargs: 键名=键值；where——key的补充。
        """
        self.key_and_table_is_exists(self.get_real_table_name(table), where_key, **kwargs)

        command = f"DELETE FROM `{self.get_real_table_name(table)}` WHERE {where_key}='{where_value}'  "
        for k, v in kwargs.items():
            command += f"{k}='{v}'"

        return self._write_db(command)

    def _alter(self, table, command: str):
        """向已有表中插入键

        语法：
            ALTER TABLE `{self.get_real_table_name(table)}` ADD COLUMN
                (`device_ip` VARCHAR(32) DEFAULT NULL COMMENT '设备IP',
                 `device_name` VARCHAR(128) DEFAULT NULL COMMENT '设备名称',
                 );
        :return:
        """
        if command.strip().endswith(','):
            command = command.strip()[:-1] + ' '
        _c = (f"ALTER TABLE `{self.get_real_table_name(table)}` ADD COLUMN ( "
              + command +
              ");")
        return self._write_db(_c)

    def write_db(self, command, *args):
        """write_db的外部访问"""
        return self._write_db(command, *args)

    def write_rows(self, c, *args):
        """write_rows的外部访问"""
        return self._write_affair(c, *args)

    def read_db(self, command, args=None, result_type=None):
        """读取数据库的外部访问"""
        return self._read_db(command, args, result_type)

    def show_tables(self):
        """列出当前数据库的数据表"""
        return self.tables_name()

    def show_columns(self, table_name, result_type=None):
        """列出指定表的字段名

        :param table_name:
        :param result_type: None | [list|tuple] | dict
        :return:
        """
        table_name = self.get_real_table_name(table_name)
        if result_type is None:
            return self.columns_name(table_name)
        elif result_type in [list, tuple]:
            return self._columns(table_name, result_type=list)
        else:
            return self._columns(table_name)

    def test_show(self):
        """数据库链接检测"""
        return self._read_db('show tables')


class MySqlAPI(MyBaseSQL, BaseSQLAPI):
    """

    **kwargs in _init_() : Optionally

        :param bind_address:
                当本地客户端有多个IP地址时，指定一个主机名或者IP。
                    When the client has multiple network interfaces, specify
                    the interface from which to connect to the host. Argument can be
                    a hostname or an IP address.
        :param unix_socket:
                可选的，您可以使用unix套接字而不是TCP/IP。
                    Optionally, you can use a unix socket rather than TCP/IP.
        :param read_timeout:
                读取连接的超时(以秒为单位)(默认:None - no timeout)
                    The timeout for reading from the connection in seconds (default: None - no timeout)
        :param write_timeout:
                写入连接的超时(以秒为单位)(默认:None - no timeout)
                    The timeout for writing to the connection in seconds (default: None - no timeout)
        :param sql_mode:
                我不太明白！
                    Default SQL_MODE to use.
        :param read_default_file:
                指定my.cnf文件从[client]部分读取这些参数。
                    Specifies  my.cnf file to read these parameters from under the [client] section.
        :param conv:
                转换字典以代替默认字典使用。这用于提供类型的自定义编组和解组。
                    Conversion dictionary to use instead of the default one.
                    This is used to provide custom marshalling and unmarshaling of types.
                    See converters.
        :param use_unicode:
                是否默认为unicode字符串。对于Py3k，这个选项默认为true。
                    Whether or not to default to unicode strings.
                    This option defaults to true for Py3k.
        :param client_flag:
                要发送到MySQL的自定义标记。在constants.CLIENT中找到潜在的价值。
                    Custom flags to send to mysql. Find potential values in constants.CLIENT.
        :param cursorclass:
                要使用的自定义游标类。 - 创建的数据库链接句柄默认的游标类。
                    Custom cursor class to use.
        :param init_command:
                建立连接时要运行的初始SQL语句。
                    Initial SQL statement to run when connection is established.
        :param connect_timeout:
                连接时抛出异常之前的超时。
                    Timeout before throwing an exception when connecting.
                        (default: 10, min: 1, max: 31536000)
        :param ssl:
                与mysql_ssl_set()的参数类似的一组参数。
                    A dict of arguments similar to mysql_ssl_set()'s parameters.
        :param read_default_group:
                要从配置文件中读取的组。
                    Group to read from in the configuration file.
        :param autocommit:
                自动提交模式。None表示使用服务器默认值。(默认值:False)
                    Autocommit mode. None means use server_test default. (default: False)
        :param local_infile:
                Bool，允许使用加载数据本地命令。(默认值:False)
                    Boolean to enable the use of LOAD DATA LOCAL command. (default: False)
        :param max_allowed_packet:
                以字节为单位发送到服务器的包的最大大小。(默认值:16 mb)
                    Max size of packet sent to server_test in bytes. (default: 16MB)
                        Only used to limit size of "LOAD LOCAL INFILE" data packet smaller than default (16KB).
        :param defer_connect:
                不要显式连接在建设-等待连接调用。 (默认值:False)
                    Don't explicitly connect on contruction - wait for connect call. (default: False)
        :param auth_plugin_map:
                处理该插件的类的插件名的字典。该类将把Connection对象作为构造函数的参数。
                该类需要一个身份验证方法，该方法将身份验证包作为一个参数。
                对于dialog插件，可以使用提示(echo, prompt)方法，(如果没有验证方法)用于从用户返回字符串。(实验)。
                    A dict of plugin names to a class that processes that plugin.
                    The class will take the Connection object as the argument to the constructor.
                    The class needs an authenticate method taking an authentication packet as
                    an argument.  For the dialog plugin, a prompt(echo, prompt) method can be used
                    (if no authenticate method) for returning a string from the user. (experimental)
        :param server_public_key:
                SHA256 用户鉴权 plugin公钥值。(默认值:无)。
                    SHA256 authenticaiton plugin public key value. (default: None)
        :param db:
                数据库的别名。(为了兼容MySQLdb)。
                    Alias for database. (for compatibility to MySQLdb)
        :param passwd:
                别名密码。(为了兼容MySQLdb)。
                    Alias for password. (for compatibility to MySQLdb)
        :param binary_prefix:
                在字节和字节数组上添加_binary前缀。(默认值:False)。
                    Add _binary prefix on bytes and bytearray. (default: False)

        See `Connection <https://www.python.org/dev/peps/pep-0249/#connection-objects>`_ in the
        specification.
        """

    def __init__(self, host, port, user, passwd, db, charset='utf8', warning=True, **kwargs):
        super().__init__(host, port, user, passwd, db, charset, **kwargs)
        if not warning:
            filterwarnings("ignore", category=pymysql.Warning)

    def create_table_compatible(self, cmd):
        return cmd

    def show_dbs(self):
        pass


class MyMySqlAPI(MySqlAPI):
    """API别名"""


if __name__ == '__main__':
    logger.addHandler(terminal_handler)  # 添加控制台



#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""
@File Name  : base
@Author     : LeeCQ
@Date-Time  : 2021/1/8 20:46
"""
from abc import ABC, abstractmethod

__all__ = ['APIBase', 'DBBase']


class DBBase(ABC):
    """基类"""

    # TABLE_PREFIX = ''
    _sql = '__sql_connect()'

    @staticmethod
    def __sql_connect():
        return ''

    @abstractmethod
    def _write_db(self, command, args=None):
        """单条数据库的写入"""

    @abstractmethod
    def _write_affair(self, command, args):
        """事务"""

    def _read_db(self, command, args=None, result_type=None):
        """读数据库"""

    @abstractmethod
    def close(self):
        """关闭数据库连接句柄"""

    def set_prefix(self, prefix):
        """设置表前缀"""
        self.TABLE_PREFIX = prefix

    def parse_prefix(self, name):
        """返回一个正确的真实的表名称"""
        return name if name.startswith(self.TABLE_PREFIX) else self.TABLE_PREFIX + name

    def get_real_table_name(self, name):
        return self.parse_prefix(name)


class APIBase(ABC):
    """实现的抽象基类"""

    @abstractmethod
    def show_dbs(self, *args, **kwargs):
        """输出数据库"""

    @abstractmethod
    def show_tables(self, *args, **kwargs):
        """展示数据表"""

    @abstractmethod
    def insert(self, *args, **kwargs):
        """插入数据（增）"""

    @abstractmethod
    def delete(self, *args, **kwargs):
        """删除数据行（删）"""

    @abstractmethod
    def update(self, *args, **kwargs):
        """更新数据（改）"""

    @abstractmethod
    def select(self, *args, **kwargs):
        """查询数据（查）"""

    @abstractmethod
    def drop_table(self, *args, **kwargs):
        """删除数据表"""

    @abstractmethod
    def drop_db(self, *args, **kwargs):
        """删除数据库"""

    @abstractmethod
    def create_table(self, *args, **kwargs):
        """创建数据表"""

    @abstractmethod
    def alter_table(self, *args, **kwargs):
        """修改数据表"""



#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""
@File Name  : error
@Author     : LeeCQ
@Date-Time  : 2021/1/8 22:45
"""
try:
    from pymysql.err import (Error,
                             DataError,
                             DatabaseError,
                             OperationalError,
                             ProgrammingError,
                             InternalError,
                             InterfaceError,
                             IntegrityError,
                             NotSupportedError,
                             MySQLError,
                             )
    from sqlite3 import (Error,
                         DataError,
                         DatabaseError,
                         OperationalError,
                         ProgrammingError,
                         InternalError,
                         InterfaceError,
                         IntegrityError,
                         NotSupportedError,
                         )
except:
    pass


class SqllibError(Exception):
    """sqllib Error"""


class InsertZipError(SqllibError):
    """批量插入数据时，各数据的长度不一致"""


class SqlKeyNameError(SqllibError):
    pass


class SqlTableNameError(SqllibError):
    pass


class SqlWriteError(SqllibError):
    pass


class SqlModuleError(SqllibError):
    pass



#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""
@File Name  : sql_base
@Author     : LeeCQ
@Date-Time  : 2021/1/8 20:46
"""
from abc import ABC, ABCMeta, abstractmethod
from warnings import warn

from .base import DBBase, APIBase
from .error import *

__all__ = ['BaseSQL', 'BaseSQLAPI']

from .common import sql_join


# from sqllib.SQLite.sqlite import SQLiteBase


class BaseSQL(DBBase, ABC):
    """关系型数据库的基类"""

    SQL_DB = None
    # 数据库

    @abstractmethod
    def tables_name(self):
        pass

    @abstractmethod
    def columns_name(self, table):
        pass

    @staticmethod  # 插入或更新多条数据时，数据格式转换
    def zip_data_for_insert(values):
        """一次插入数据库多条数据时，打包相应的数据。"""
        for x in values:
            if not isinstance(x, (tuple, list)):
                raise InsertZipError(f"INSERT多条数据时，出现非列表列！确保数据都是list或者tuple。\n错误的值是：{x}")

            if not len(values[0]) == len(x):
                raise InsertZipError(f'INSERT多条数据时，元组长度不整齐！请确保所有列的长度一致！\n'
                                     f'[0号]{len(values[0])}-[{values.index(x)}号]{len(x)}')

        return tuple([v for v in zip(*values)])  # important *

    # 判断表、键值的存在性
    def key_and_table_is_exists(self, table, key, *args, **kwargs):
        """ 判断 key & table 是否存在

        :param table: 前缀 + 表单名
        :param key: 键名
        :param args: 键名, 多个键名
        :param kwargs: 键名=键值；
        :return: 0 存在
        """
        tables_name = [i.upper() for i in self.tables_name()]
        if table.upper() not in tables_name:
            raise SqlTableNameError(f"{table} NOT in This Database: {self.SQL_DB};\n"
                                    f"(ALL Tables {self.tables_name()}")

        cols = [i.decode() if isinstance(i, bytes) else i for i in self.columns_name(table)]
        not_in_table_keys = [k.upper() for k, v in kwargs.items() if k not in cols]
        not_in_table_keys += [k.upper() for k in args if k not in cols]
        if key.upper() not in cols and not_in_table_keys:
            raise SqlKeyNameError(f'The key {key.upper()} NOT in this Table: {table};\n'
                                  f'(ALL Columns {cols}) \n'
                                  f'self.columns_name(table) = {self.columns_name(table)}'
                                  )
        return 0

    @abstractmethod
    def _create_table(self, cmd, table_name, exists_ok, table_args, *args):
        pass

    @abstractmethod
    def _insert(self, table, ignore_repeat=False, **kwargs):
        pass

    @abstractmethod
    def _select(self, table, cols, *args, result_type=None, **kwargs):
        pass

    @abstractmethod
    def _update(self, table, where_key, where_value, **kwargs):
        pass

    @abstractmethod
    def _drop(self, option, name):
        pass

    @abstractmethod
    def _delete(self, table, where_key, where_value, **kwargs):
        pass

    @abstractmethod
    def _alter(self, table, command: str):
        pass


class BaseSQLAPI(BaseSQL, APIBase, metaclass=ABCMeta):

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __enter__(self):
        return self

    @abstractmethod
    def create_table_compatible(self, cmd):
        """对CREATE语句中兼容的语法进行重置或删除"""
        return cmd

    def create_table(self, cmd: (str, tuple), table_name, exists_ok=False, table_args='', *args):
        """ 创建一个数据表：

        模板：
            _c = (f"CREATE TABLE IF NOT EXISTS `{table_name}` ( "
                    f"a VARCHAR(10),"
                    f"b VARCHAR(10)"
                    f" ) ")
        这样就可以创建一个名为'table_name' 的数据表；
        有2个键：a, b 都是变长字符串(10)

        :param table_name:
        :param exists_ok: 为False，创建的表存在时，抛出异常
        :param cmd 字段字符串
        :param table_args: 创建表时附加的额外的参数 CREATE TABLE `NAME` ( params ) >>table_args<< ;
        :param args: 无，不要填
        :return: 0 成功
        """
        if isinstance(cmd, (tuple, list)):
            cmd = sql_join(cmd)[0]
        cmd = self.create_table_compatible(cmd)
        table_name = self.get_real_table_name(table_name)
        return self._create_table(cmd, table_name, exists_ok=exists_ok, table_args=table_args, *args)

    def insert(self, table, ignore_repeat=False, **kwargs):
        """ 向数据库插入内容。

        允许一次插入多条数据，以 key=(tuple)的形式；
            但是要注意，所有字段的元组长度需要相等；否组报错。

        :param table: 表名；
        :param ignore_repeat: 忽视重复
        :param kwargs: 字段名 = 值；字段名一定要存在与表中， 否则报错；
        :return: 0 成功 否则 报错
        """
        return self._insert(table, ignore_repeat=ignore_repeat, **kwargs)

    def select(self, table, cols, *args, result_type=None, **kwargs):
        """ 从数据库中查找数据；

            column_name 可以设置别名；

            注意： `就不再支持 * `
            注意：column_name, table, key 可以用 ` ` 包裹， value 一定不能用，
                  如果有需要，value 用 ' ' 。
        :param table:
        :param cols: 传参时自行使用 `` , 尤其是数字开头的参数
        :param result_type: 返回结果集：{dict, None, tuple, 'SSCursor', 'SSDictCursor'}
        :param kwargs: {'WHERE', 'LIMIT', 'OFFSET', 'ORDER'} 全大写
                        WHERE 查询字符串 如 KEY=VALUE
                        LIMIT 2 OFFSET 2  通常连在一起使用
                        ORDER 排序  COL_NAME  [ASC | DESC]
                      特殊键：result_type = {dict, None, tuple, 'SSCursor', 'SSDictCursor'}
        :return 结果集 通过键 - result_type 来确定 -
        """

        _cols = []
        if isinstance(cols, str):
            _cols.append(cols)
        if isinstance(cols, (list, tuple)):
            [_cols.append(_) for _ in cols]
        if args:
            [_cols.append(_) for _ in args]
        return self._select(table, _cols, result_type=result_type, **kwargs)

    def select_new(self, table, columns_name: tuple or list, result_type=None, **kwargs):
        """ SELECT的另一种传参方式：
                要求所有的查询字段放在一个列表中传入。

        :param table:
        :param columns_name:
        :type columns_name: tuple or list
        :param result_type: 返回结果集类型：{dict, None, tuple, 'SSCursor', 'SSDictCursor'}
        :return: 结果集 通过键 - result_type 来确定
        """
        warn('不建议使用此函数！')
        return self.select(table, columns_name, result_type=result_type, **kwargs)

    def update(self, table, where_key, where_value, **kwargs):
        """ 更新数据库数据：

        :rtype: int
        :param table: 表名
        :param where_key: 通过字段查找行的键
        :param where_value: 其值
        :param kwargs: 需要更新的键值对
        :return: 0 or Error
        """
        return self._update(table, where_key, where_value, **kwargs)

    def drop_table(self, name):
        """用来删除一张表

        :param name: table name
        :return: 0 or Error
        """
        return self._drop('TABLE', name)

    def drop_db(self, name):
        """用来删除一个数据库

        :param name:
        :return:
        """
        return self._drop('DB', name)

    def delete(self, table, where_key, where_value, **kwargs):
        """ 用来删除数据表中的一行数据；

        :param table: 表名
        :param where_key: 查找到键
        :param where_value: 查找的值 可以使用表达式
        :param kwargs: 补充查找的键值对；
        :return: 0 or Error
        """
        return self._delete(table, where_key=where_key, where_value=where_value, **kwargs)

    def alter_table(self, table, command: str):
        """向已有表中插入键

        语法：
            ALTER TABLE `{self.TABLE_PREFIX}{table_name}` ADD COLUMN
                (`device_ip` VARCHAR(32) DEFAULT NULL COMMENT '设备IP',
                 `device_name` VARCHAR(128) DEFAULT NULL COMMENT '设备名称',
                 );
        :return:
        """
        return self._alter(table, command)



# -*- coding: utf-8 -*-
"""
@Time   : 2019/11/25 13:15
@Author : LeeCQ

[mysql(pymysql, 需要服务器),  MongoDB(pyMongoDB，需要服务器),
 SQLite(关系型数据库, Python原生支持),
 bsddb(NoSQL Bsddb3),
 UnQLite(NoSQL),
 LiteDB(.NET开发轻量级,文件数据库, For Python v1.0.0),  # https://pypi.org/project/litedb/
 CodernityDB(纯Python开发, 功能不足),  #
 Redis(需要安装软件),


]

mysql:

SQLite:

MongoDB:

bsddb: ??

UnQLite:
    描述：嵌入式NoSQL - 体积小，无服务器， 源码付费
    官网：https://unqlite.org/
    简介：https://blog.csdn.net/robinL2005/article/details/40182699

LiteDB:
    官网：http://www.litedb.org/
"""

from .SQLite import SQLiteAPI
from .mysql import MyMySqlAPI, MySqlAPI
from .common.common import sql_join
from .common.base_sql import BaseSQL, BaseSQLAPI
from .common import common

# 直接访问会出错，但是，其他模块可以正常导入这些API
__version__ = '0.2.6.2'

# __all__ = ['common', 'BaseSQL',
#            'SQLiteAPI', 'MyMySqlAPI', 'MySqlAPI', 'sql_join']



#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""
Setup module for core.

@File Name  : setup.py
@Author     : LeeCQ
@Date-Time  : 2020/7/30 11:42
"""
import os
from setuptools import setup, find_packages

PACKAGE = "sqllib"
DESCRIPTION = "对MySQL, SQLite数据库进行接口集成."
AUTHOR = "Lee CQ"
AUTHOR_EMAIL = "lee-cq@qq.com"
URL = "https://leecq.cn"

TOPDIR = os.path.dirname(__file__) or "."
VERSION = __import__(PACKAGE).__version__

with open("readme.md", encoding='utf8') as fp:
    LONG_DESCRIPTION = fp.read()


with open("requirements.txt") as fp:
    requires = fp.read().split('\n')

with open('./updateLog', encoding='utf8') as fp:
    update_log = fp.read()


setup_args = {
    'version': VERSION,
    'description': DESCRIPTION,
    'long_description': LONG_DESCRIPTION,
    'author': AUTHOR,
    'author_email': AUTHOR_EMAIL,
    'license': "Apache License 2.0",
    'url': URL,
    'keywords': ["mysql", "API", "SQLite", 'DB'],
    'packages': find_packages(exclude=["tests*"]),
    # 'package_data': {'aliyunsdkcore': ['data/*.json', '*.pem', "vendored/*.pem"],
    #                  'aliyunsdkcore.vendored.requests.packages.certifi': ['cacert.pem']},
    'platforms': 'any',
    # 'install_requires': requires,
    'classifiers': [
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Software Development',
    ]
}

setup(name=PACKAGE, **setup_args)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""
@File Name  : sqlite.py
@Author     : LeeCQ
@Date-Time  : 2020/1/13 12:30

SQLite

2020.01.28 -- _select OFFSET 添加str()
"""
import logging
import sys
import sqlite3
import re
from sqllib.common.common import sql_join
from sqllib.common.base_sql import BaseSQL, BaseSQLAPI
from sqllib.common.error import *

logger = logging.getLogger("logger")  # 创建实例
formatter = logging.Formatter("[%(asctime)s] < %(funcName)s: %(thread)d > [%(levelname)s] %(message)s")
# 终端日志
console_handler = logging.StreamHandler(sys.stdout)
console_handler.setFormatter(formatter)  # 日志文件的格式
logger.setLevel(logging.DEBUG)  # 设置日志文件等级

__all__ = ['SQLiteAPI', 'SQLiteBase']


# 以字典形式返回游标的sqlite实现
def dict_factory(cursor, row) -> dict:
    """用来替换sqlite.connect().row_factory

    :return: dict
    """
    d = {}
    for idx, col in enumerate(cursor.description):
        d[col[0]] = row[idx]
    return d


class SQLiteBase(BaseSQL):
    """SQLite实现的基类

    """

    def __init__(self, db, **kwargs):
        self._sql = sqlite3.connect(db, **kwargs)
        self.TABLE_PREFIX = '' if 'prefix' not in kwargs.keys() else kwargs['prefix']

    def close(self):
        self._sql.close()

    def close_db(self):
        """关闭数据库链接，并退出"""
        self._sql.close()

    def pooling_sql(self):
        pass

    @property
    def get_connect(self):
        """返回数据库链接句柄"""
        return self._sql

    # 写数据库操作
    def _write_db(self, command, args=None):
        __sql = self._sql
        cur = __sql.cursor()  # 使用cursor()方法获取操作游标
        try:
            if not args:
                cur.execute(command)
            else:
                cur.execute(command, args)
            __sql.commit()  # 提交数据库
            return cur.rowcount
        except Exception as e:
            __sql.rollback()
            raise SqlWriteError(f'操作数据库时出现问题，数据库已回滚至操作前：{e}'
                                f'\n\n>>COMMEND:\n{command}\n>>ARGS:\n{args}')
        finally:
            cur.close()

    def __write_no_except(self, command, args=None):
        """未收集错误的"""
        __sql = self._sql
        cur = __sql.cursor()  # 使用cursor()方法获取操作游标
        if not args:
            cur.execute(command)
        else:
            cur.execute(command, args)
        __sql.commit()  # 提交数据库
        cur.close()
        return 0

    def _write_affair(self, command, args):
        """数据库事务写入。"""
        __sql = self._sql
        cur = __sql.cursor()
        try:
            cur.executemany(command, args)
            __sql.commit()
            return cur.rowcount
        except Exception as e:
            __sql.rollback()
            sys.exc_info()
            raise SqlWriteError(f"执行写事物时出错，已回滚：{e}"
                                f"\n\n>>COMMEND:\n{command}\n>>ARGS:\n{args}")
        finally:
            cur.close()

    def __read_db(self, command, result_type=None):
        """数据库读取的具体实现。主要涉及数据库查询"""
        __sql = self._sql
        if result_type is dict:
            __sql.row_factory = dict_factory
        else:
            __sql.row_factory = sqlite3.connect('').row_factory
        cur = __sql.cursor()
        cur.execute(command)
        results = cur.fetchall()
        cur.close()
        return results

    def read_db(self, cmd, result_type=None):
        """是菊科操作的对外接口。"""
        return self.__read_db(cmd, result_type=result_type)

    def write_db(self, cmd, args=None):
        """数据库操作的对外接口。"""
        # logger.warning("使用此函数时注意表前缀! ")
        return self._write_db(cmd, args)

    def write_no_except(self, cmd, args=None):
        """数据库写入对外接口，它没有收集任何错误! """
        logger.warning("使用此函数时注意表前缀! ")
        return self.__write_no_except(cmd, args)

    # 创建数据库
    def _create_table(self, keys, table_name, exists_ok, table_args, *args):
        """创建数据表

           CREATE TABLE IF NOT EXISTS COMPANY(
           ID INT PRIMARY KEY     NOT NULL,
           NAME           TEXT    NOT NULL,
           AGE            INT     NOT NULL,
           ADDRESS        CHAR(50),
           SALARY         REAL
            );
        """
        if isinstance(keys, tuple):
            keys = sql_join(keys)[1]  #
        keys = keys.rstrip().rstrip(',')
        _ignore = ' IF NOT EXISTS ' if exists_ok else ''

        _c = f"CREATE TABLE {_ignore} {self.get_real_table_name(table_name)} ( "
        _c += keys + ");"
        logger.debug(_c)
        return self._write_db(_c)

    # 插入表数据
    def _insert(self, table_name, ignore_repeat=False, **kwargs):
        """ 向数据库插入内容。

        :param table_name: 表名；
        :param ignore_repeat: 唯一约束失败时, 忽略重复错误.
        :param kwargs: 字段名 = 值；
        :return:
        """
        _ignore = 'OR IGNORE' if ignore_repeat else ''
        _c = f"INSERT {_ignore} INTO {self.get_real_table_name(table_name)} ( "
        _c += ", ".join([_ for _ in kwargs.keys()]) + " ) "
        _c += "VALUES ( "
        _c += ", ".join([f"?" for _ in kwargs.values()]) + " ) ; "
        # print(_c, kwargs)
        if isinstance(list(kwargs.values())[0], (tuple, list)):  # kwargs第一个键是不是字符串
            arg = self.zip_data_for_insert(tuple(kwargs.values()))
            return self._write_affair(_c, arg)
        else:
            for x in kwargs.values():
                if isinstance(x, (list, tuple)):
                    raise InsertZipError(f"INSERT一条数据时，出现列表列或元组！确保数据统一: VALUE({x})")
            return self._write_db(_c, list(kwargs.values()))  # 提交

    # def _insert_rows(self, table_name, args, k=None, ignore_repeat=False):
    #     """插入
    #
    #     :param table_name: 表名
    #     :param args:
    #     :param k:
    #     :param ignore_repeat:
    #     :return:
    #     """
    #     _a = [_.values() for _ in args]
    #     if k is None:
    #         if not isinstance(args[0], dict):
    #             raise ValueError(f'既没有k, 也不是dict')
    #         k = args[0].keys()
    #     _ignore = 'OR IGNORE' if ignore_repeat else ''
    #     _c = f"INSERT {_ignore} INTO {self.get_real_table_name(table)} ( "
    #     _c += ", ".join([_ for _ in k]) + " ) "
    #     _c += "VALUES ( "
    #     _c += ", ".join([f"?" for _ in args[0].values()]) + ");"
    #     return self.__write_many(_c, _a)

    # 检索表
    def _select(self, table, cols: list and tuple, result_type=None, **kwargs):
        """ select的应用。

            ·· `就不再支持 * `
            ·· column_name, table, key 可以用 ` ` 包裹， value 一定不能用， value 用 ' ' 。
        :param result_type: {dict, other}
        :param table:
        :param cols: 传参时自行使用 `` , 尤其是数字开头的参数
        :param kwargs: {'WHERE', 'LIMIT', 'OFFSET', ORDER} 全大写
        :return 结果集
        """
        command = f'SELECT  ' + ' , '.join(cols) + ' ' + f'FROM `{self.get_real_table_name(table)}` '
        # ===========================================
        # 下面是更好的解决方案
        # ===========================================
        # for key, value in kwargs.items():
        #     key = key.upper()
        #     if key in ['WHERE', 'LIMIT', 'OFFSET']:
        #         command += f' {key}  {value}'
        #     if key == 'ORDER':
        #         command += f' {key} BY {value}'
        # ===========================================
        command += ' '.join([' '.join((key.upper(), str(value))) for key, value in kwargs.items()
                             if key.upper() in ['WHERE', 'LIMIT', 'OFFSET']]) + '  '
        command += ' '.join([f'ORDER BY {value}' for key, value in kwargs.items()
                             if key.upper() == 'ORDER']) + ' '
        # print(command)
        return self.__read_db(command, result_type=result_type)

    # 更新表
    def _update(self, table, where_key, where_value, **kwargs):
        """ 更新数据库

        :param table: 数据表名字
        :param where_key: where的键
        :param where_value: where的值
        :param kwargs: 更新的键 = 更新的值， 注意大小写，数字键要加 - ``
        :return: 0 成功。
        """

        self.key_and_table_is_exists(self.TABLE_PREFIX + table, where_key, **kwargs)  # 判断 表 & 键 的存在性！
        command = f"UPDATE `{self.get_real_table_name(table)}` SET  " + \
                  ' , '.join([f" {k}=? " for k, v in kwargs.items()]) + \
                  f" WHERE `{where_key}`='{where_value}' ;"  # 构造WHERE语句

        return self._write_db(command, [_ for _ in kwargs.values()])  # 执行SQL语句

    # 删除表数据 一行
    def _delete(self, table, where_key, where_value, **kwargs):
        """删除数据表中的某一行数据，
        :param table:
        :param kwargs: 键名=键值；最好是数据库的主键或唯一的键。如果数据库没有，则最好组合where，以保证删除 - 唯一行。
        """
        self.key_and_table_is_exists(self.get_real_table_name(table), where_key, **kwargs)  # 判断键的存在性

        command = f"DELETE FROM `{self.get_real_table_name(table)}` WHERE {where_key}='{where_value}' "
        print(kwargs)
        for k, v in kwargs.items():
            command += f'{k}="{v}"'
        return self._write_db(command)

    # 删除表或者数据库
    def _drop(self, option, name):
        """ 删除数据库内容：

        :param option: (TABLE or INDEX)
        :param name:
        :return: 0 成功
        """
        if option.upper() == 'TABLE':
            command = f'DROP  {option}  `{self.get_real_table_name(name)}`'
        else:
            command = f'DROP  {option}  `{name}`'
        return self._write_db(command)

    def tables_name(self) -> list:
        """"""
        return self.show_tables(name_only=True)

    def show_tables(self, name_only: bool = True):
        """

        :param name_only:
        :return: {'type': 类型(table), 'name': 名称, 'tbl_name': 所属表明, 'rootpage': 表编号, 'sql':创建表时执行的SQL语句
        """
        if name_only:
            return [_[0] for _ in self.read_db('SELECT name FROM sqlite_master where type="table";')]
        else:
            return [_ for _ in self.read_db('SELECT * FROM sqlite_master where type="table";', result_type=dict)]

    def columns_name(self, table):
        """"""
        return self.show_columns(table_name=table, name_only=True)

    def show_columns(self, table_name, name_only=True) -> list:
        """显示列信息

        :return: {'cid': 0, 'name': 'id', 'type': 'int', 'notnull': 0, 'dflt_value': None, 'pk': 0}
                  字段编号     字段名        字段类型        不允许null      默认值
        """
        table_name = self.get_real_table_name(table_name)
        if name_only:
            return [_['name'] for _ in self.read_db(f'PRAGMA table_info({table_name});', result_type=dict)]
        else:
            return [_ for _ in self.read_db(f'PRAGMA table_info({table_name});', result_type=dict)]

    # 修改表结构
    def _alter(self, table_name, command):
        """ 目前sqlite只支持RENAME，ADD COLUMN """
        table = self.get_real_table_name(table_name)
        _c = f'ALTER TABLE {table}  {command}'
        return self._write_db(_c)


class SQLiteAPI(SQLiteBase, BaseSQLAPI):
    """接口类"""

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close_db()

    def __enter__(self):
        return self

    def show_dbs(self, *args, **kwargs):
        pass

    # def insert(self, table_name, ignore_repeat=False, **kwargs):
    #     """插入数据到数据库，支持插入多条数据。
    #
    #     :param table_name:
    #     :param ignore_repeat:
    #     :param kwargs: 需要插入的 字段名=值 （当值为list or tuple时，开启事物插入多条数据）
    #                     当多个字段的list长度不一致时，抛出ZIPError。
    #     :return:
    #     """
    #     return self._insert(table_name, ignore_repeat=ignore_repeat, **kwargs)

    # 更好的解决方案：insert(ignore_repeat=True, ...)
    def insert_line2line(self, table_name, **kwargs):
        """已单条的形式插入大量数据 - 不推荐"""
        logger.warning('这个函数的性能极低！谨慎使用！')
        if not isinstance(list(kwargs.values())[0], str):  # kwargs第一个键是不是字符串
            arg = self.zip_data_for_insert(tuple(kwargs.values()))
            for _a in arg:  # 关键逻辑
                try:
                    self._insert(table_name, **{_k: _a[index] for index, _k in enumerate(kwargs.keys())})
                except Exception as e:
                    logger.warning(f'Except: {e}')
            return 0
        else:
            raise SqlModuleError('插入单条数据请使用 insert()方法。')

    # def select(self, table, cols, *args, result_type=None, **kwargs):
    #     """ SQLite 查询数据库
    #
    #     :param table:
    #     :param cols:
    #     :param result_type:
    #     :param kwargs: {'WHERE', 'LIMIT', 'OFFSET', ORDER} 全大写
    #     :return:
    #     """
    #     _cols = []
    #     if isinstance(cols, str):
    #         _cols.append(cols)
    #     if isinstance(cols, (list, tuple)):
    #         [_cols.append(_) for _ in cols]
    #     if not args:
    #         [_cols.append(_) for _ in args]
    #     # print(_cols)
    #     return self._select(table, _cols, result_type=result_type, **kwargs)

    # def update(self, table, where_key, where_value, **kwargs):
    #     """更新数据库数据 -
    #
    #     :param table:
    #     :param where_key:
    #     :param where_value:
    #     :param kwargs:
    #     :return:
    #     """
    #     return self._update(table, where_key, where_value, **kwargs)

    # def delete(self, table, **kwargs):
    #     """删除数据表中的某一行数据，
    #     :param table:
    #     :param kwargs: 键名=键值；最好是数据库的主键或唯一的键。如果数据库没有，则最好组合where，以保证删除 - 唯一行。
    #     :return
    #     """
    #     return self._delete(table, **kwargs)

    # def drop_table(self, table_name):
    #     """删除一个数据库"""
    #     return self._drop('TABLE', table_name)

    def drop_index(self, index_name):
        """删除一个索引"""
        return self._drop('INDEX', index_name)

    def alter_rename(self, old_name, new_name):
        """重命名表"""
        new_name = self.TABLE_PREFIX + new_name
        cmd = f'RENAME TO `{new_name}`'
        return self._alter(old_name, cmd)

    def alter_add_column(self, table_name, columns_info):
        """怎加表的列"""
        cmd = f'ADD COLUMN {columns_info}'
        return self._alter(table_name, cmd)

    def create_table_compatible(self, cmd: str):
        # 移除COMMENT语句
        cmd = re.sub(r'(COMMENT +".*?")', '', cmd, flags=re.IGNORECASE)
        # MySQL自增 --> SQLite r'[(, ]* *(\w*).*?AUTO_INCREMENT.*?,'
        for _r in re.findall(r'[(, ]* *(\w*).*?AUTO_INCREMENT.*?,', cmd, re.IGNORECASE):
            cmd = re.sub(r'[(, ]* *\w*.*AUTO_INCREMENT.*?,', f'{_r} INTEGER PRIMARY KEY AUTOINCREMENT , ', cmd)

        _integer = {'INT', 'TINYINT', 'SMALLINT', 'MEDIUMINT', 'BIGINT', 'UNSIGNED BIG INT', 'INT2', 'INT8', }
        _real = {'DOUBLE', 'FLOAT', 'DOUBLE PRECISION', 'TIMESTAMP'}
        _text = {'CHARACTER', 'VARCHAR', 'VARYING CHARACTER', 'NCHAR', 'NATIVE CHARACTER', 'NVARCHAR', 'TEXT', 'CLOB',
                 'CHAR',
                 }
        _blob = {'BLOB', 'TINYBLOB', 'MEDIUMBLOB', 'LONGBLOB'}

        cmd = re.sub(' ' + f'({" | ".join(_integer)})' + ' ', ' integer ', cmd, flags=re.IGNORECASE)
        cmd = re.sub(' ' + f'({" | ".join(_real)})' + ' ', ' REAL ', cmd, flags=re.IGNORECASE)
        cmd = re.sub(' ' + f'({" | ".join(_text)})' + ' ', ' TEXT ', cmd, flags=re.IGNORECASE)
        cmd = re.sub(' ' + f'({" | ".join(_blob)})' + ' ', ' BLOB ', cmd, flags=re.IGNORECASE)
        return cmd


if __name__ == '__main__':
    logger.addHandler(console_handler)  #
    table_structure = (  # f' CREATE TABLE `{table_name}` ( '
        f'_ID INT AUTO_INCREMENT PRIMARY KEY , '
        f'TEST_STR VARCHAR(100) NOT NULL COMMENT "随机字符串", '
        f'TEST_DATETIME DATETIME NOT NULL COMMENT "当前时间", '
        f'TEST_INT INT NOT NULL COMMENT "整数时间戳", '
        f'TEST_FLOAT DOUBLE NOT NULL COMMENT "时间戳" , '
        f'TEST_JSON JSON NOT NULL COMMENT "测试JSON格式写入" , '
        f'TEST_BLOB MEDIUMBLOB COMMENT "测试二进制数据的写入"'
    )
    a = SQLiteAPI('')
    a.create_table_compatible(table_structure)



#!/usr/bin/python3 
# -*- coding: utf-8 -*-
"""
@File Name  : test_1.py
@Author     : LeeCQ
@Date-Time  : 2020/1/19 22:00

"""
import json
# t1 = json.load(open(r'C:\code\Py\HTTP\Download_9C\data\1.txt'))
import pymongo
client = pymongo.MongoClient("mongodb+srv://root:root123456@test-ivnj5.azure.mongodb.net/test?retryWrites=true&w=majority")
db = client.test
collect = db['9c']





#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""
@File Name  : test_sql.py
@Author     : LeeCQ
@Date-Time  : 2020/7/14 19:00

0. 连接检查
1. 创建数据库
2. 插入数据
3. 查询数据
4. 修改数据（update）
5. 删除数据（行）
6. 修改表结构
7. drop操作
9. 删除表操作
"""
import unittest
import json
from time import time
# from datetime import n
from pathlib import Path
from sqllib.mysql.mysqlbase import MySqlAPI
from sqllib.SQLite.sqlite import SQLiteAPI
from sqllib.common.base_sql import BaseSQL
from sqllib.common.error import *

__DEBUG__ = True
__SQLite__ = 'sup/UT_SQLite.sqlite'  # ':memory:' if not False else
# __MySQL__ = ('t.sql.leecq.cn', 10080, 'test', 'test123456', 'test')
__MySQL__ = ('localhost', 3306, 'test', 'test123456', 'test')

table_name = 'PYTHON_UNITTEST'
table_prefix = 'UT_'
table_structure = (  # f' CREATE TABLE `{table_name}` ( '
    f'_ID INT AUTO_INCREMENT PRIMARY KEY , '
    f'TEST_STR VARCHAR(100) NOT NULL COMMENT "随机字符串", '
    f'TEST_DATETIME DATETIME NOT NULL COMMENT "当前时间", '
    f'TEST_INT INT NOT NULL COMMENT "整数时间戳", '
    f'TEST_FLOAT DOUBLE NOT NULL COMMENT "时间戳" , '
    f'TEST_JSON JSON NOT NULL COMMENT "测试JSON格式写入" , '
    f'TEST_BLOB MEDIUMBLOB COMMENT "测试二进制数据的写入"'
)
table_keys = ['TEST_STR', 'TEST_DATETIME', 'TEST_INT', 'TEST_FLOAT', 'TEST_JSON', 'TEST_BLOB']
table_data_one = (r'!@#$%^&*()_+-=/{}[]\' ";,.<>/`·This is a character test string。7894561320',
                  '2021-1-9 17:28:16', int(time()), time(),
                  Path('sup/json_data.json').read_text(encoding='utf8'),
                  Path('sup/photo.jpg').read_bytes()
                  )
table_data_more = (table_data_one,
                   ('A', '2021-1-9 18:21:55', int(time()), time(), json.dumps({"A": 1}), b'This is BLOB for MEDIUM 1 '),
                   ('B', '2021-1-9 18:24:35', int(time()), time(), json.dumps({"B": 2}), b'This is BLOB for MEDIUM 2 ')
                   )


def _dict_for_data(is_more=False):
    """打包"""
    if is_more:
        return {_x: _y for _x, _y in zip(table_keys, zip(*table_data_more))}
    else:
        return {_x: _y for _x, _y in zip(table_keys, table_data_one)}


def _zip_to_list(_zip):
    """行转列"""
    return [_ for _ in zip(*_zip)]


class TESTMySql(unittest.TestCase):

    def setUp(self) -> None:
        # self.sql = MySqlAPI(*__MySQL__, warning=False, charset='utf8')
        self.sql = SQLiteAPI(__SQLite__)
        self.sql.set_prefix(table_prefix)

    def tearDown(self) -> None:
        self.sql.close()

    def test_00_connect(self):
        self.assertIsInstance(self.sql, BaseSQL, "数据库连接失败")

    def test_01_set_pr(self):
        self.assertEqual(table_prefix, self.sql.TABLE_PREFIX, "设置表前缀未生效！")

    def test_02_pool_connect(self):
        """连接池测试"""
        self.sql.pooling_sql()

    @unittest.expectedFailure
    def test_11_create_table_noName(self):
        try:
            self.sql.drop_table(table_name)
        except:
            pass
        _cmd = f'CREATE TABLE `{table_name}` ' + ' ( ' + table_structure + ' ) '
        print(_cmd)
        self.sql.create_table(_cmd, table_name)

    def test_12_create_table_with_name(self):
        self.sql.create_table(cmd=table_structure, table_name=table_name, exists_ok=True)
        self.sql.show_tables()

    def test_13_create_table_existed(self):
        self.sql.create_table(cmd=table_structure, table_name=table_name, exists_ok=True)
        self.assertRaises(SqlWriteError, self.sql.create_table, table_structure, table_name)

    def test_14_create_table_json_sql(self):
        """join_sql()函数的使用情况"""
        _cmd = (('_ID', ' INT', 'auto_increment PRIMARY KEY '),
                ('TEST_STR', ' VARCHAR(100)', ' NOT NULL COMMENT "随机字符串"'),
                ('TEST_DATETIME', ' DATETIME NOT NULL COMMENT "当前时间"'),
                ('TEST_INT', ' INT', ' NOT NULL COMMENT "整数时间戳" '),
                ('TEST_FLOAT', ' DOUBLE', ' NOT NULL COMMENT "时间戳" '),
                ('TEST_JSON', ' JSON', ' NOT NULL COMMENT "测试JSON格式写入"  '),
                ('TEST_BLOB', ' MEDIUMBLOB', ' COMMENT "测试二进制数据的写入"'),
                )
        self.sql.create_table(_cmd, 'test_join_sql')
        self.sql.drop_table('test_join_sql')

    def test_21_insert_one(self):
        """插入一条数据"""
        _a = self.sql.insert(table_name, **_dict_for_data())
        self.assertTrue(_a > 0, f'插入的行数为 {_a}')

    def test_22_insert_more(self):
        """插入多条数据"""
        _a = self.sql.insert(table_name, ignore_repeat=False, **_dict_for_data(is_more=True))
        self.assertTrue(_a > 0, f'插入的行数为 {_a}')

    def test_23_insert_zip_error(self):
        """插入插入多条数据时，数据长度不一致"""
        _dict_data = _dict_for_data(is_more=True)
        _dict_data[table_keys[0]] = ('A', 'B')
        self.assertRaises(InsertZipError, self.sql.insert, table_name, ignore=False, **_dict_data)

    def test_24_insert_not_null_error(self):
        """NOT NULL 但是没有给值！"""
        _dict_data = _dict_for_data(is_more=False)
        _dict_data[table_keys[0]] = None
        self.assertRaises(SqlWriteError, self.sql.insert, table_name, **_dict_data)

    def test_31_select(self):
        """基本操作"""
        # print(self.sql.TABLE_PREFIX)
        _a = self.sql.select(table_name, '*')

    def test_32_select_col(self):
        """查询指定列以及传列方式的兼容性"""
        _keys = table_keys[1:3]
        _a = self.sql.select(table_name, _keys, result_type=dict)
        _b = self.sql.select(table_name, *_keys, result_type=dict)
        self.assertEqual(_a, _b, f"列表传参和*args传参返回数据不一致")

    def test_33_select_where(self):
        """查询指定行"""
        _ = [i[0] for i in self.sql.select(table_name, '_ID', where=f'{table_keys[1]}="{table_data_one[1]}"')]
        self.assertEqual([1, 2], _)

    def test_34_select_limit(self):
        """查询2个条目(2,3)"""
        _test_map = [((2, 2), [3, 4], ''),
                     ((3, 1), [2, 3, 4], ''),
                     ((1, 4), [], ''),
                     ]
        for _in, _out, msg in _test_map:
            with self.subTest(_in=_in, _out=_out, msg=msg):
                _ = [i[0] for i in self.sql.select(table_name, '_ID', LIMIT=_in[0], OFFSET=_in[1])]
                self.assertEqual(_out, _, msg)

    def test_35_select_order(self):
        """结果排序"""
        _ = [i[0] for i in self.sql.select(table_name, '_ID', ORDER=f'{table_keys[0]} DESC')]
        self.assertEqual([4, 3, 1, 2], _, )

    def test_36_show_tables(self):
        """返回表名字"""
        _ = [i.upper() for i in self.sql.tables_name()]
        print(_)
        self.assertIn(self.sql.get_real_table_name(table_name).upper(), _, '返回的数据表名称不正确')

    def test_37_show_columns(self):
        _ = self.sql.columns_name(table_name)
        print(_)
        self.assertTrue(set(table_keys).issubset(_), '返回的表字段异常')

    def test_41_update(self):
        """修改数据（更新数据）"""
        self.sql.update(table_name, '_ID', '1', TEST_STR='NEW1')
        _ = self.sql.select(table_name, 'TEST_STR', WHERE='_ID=1')[0][0]
        self.assertEqual('NEW1', _.decode() if isinstance(_, bytes) else _, "数据更改失败")

    def test_51_delete_row(self):
        """删除匹配的行"""
        self.sql.delete(table_name, '_ID', '1', )
        _ = self.sql.select(table_name, '_ID', WHERE='_ID=1')
        self.assertEqual(0, len(_))

    def test_61_alter_add_col(self):
        """修改表结构-添加列"""

    def test_62_alter_edit_col(self):
        """修改表结构-编辑列"""

    def test_63_alter_drop_col(self):
        """修改表结构-删除列"""

    @unittest.skipIf(__DEBUG__, "DEBUG-ING ...")
    def test_91_drop_table(self):
        """删除数据表测试"""
        _ = self.sql.drop_table(table_name)
        self.assertEqual(_, 0, '删除数据表失败')


class TESTSQLite(TESTMySql, unittest.TestCase):

    def setUp(self) -> None:
        self.sql = SQLiteAPI(__SQLite__)
        self.sql.set_prefix(table_prefix)


if __name__ == '__main__':
    unittest.main()
    # mysql = TESTMySql()
    # # sqlite = TESTSQLite()
    # test = unittest.TestSuite()
    # test.addTest(mysql)
    #
    # runner = unittest.TestRunner()
    # runner.run(test)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""
@File Name  : __init__.py
@Author     : LeeCQ
@Date-Time  : 2021/1/8 19:55

模块对应关系：
1. DBBase, APIBase
2. BaseSQL(DBBase)  ,  NoSQLBase(DBBase)
3.1 MyBaseSQL(BaseSQL, APIBase)  ,  SQLiteBASE(BaseSQL, APIBase)
3.2 MongdbBASE(NoSQLBase, APIBase)  ,

"""
